using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

namespace Smdn.Protocols.Imap4.Protocol {
  public class ImapProtocolResponseParser {
    public ImapProtocolResponse[] Parse(byte[] responseBytes)
    {
      var responses = new List<ImapProtocolResponse>();

      // 受信したものを全てバッファにため込む
      buffer += latin1Encoding.GetString(responseBytes);

      try {
        for (;;) {
          string unparsed;

          var response = TryParse(buffer, out unparsed);

          // 解析されなかった分をバッファに戻す
          buffer = unparsed;

          if (response == null) {
            // 解析できなかったので次のレスポンスを待つ(レスポンスの途中の場合)
            break;
          }
          else {
            responses.Add(response);

            // バッファに残っている場合は、継続して解析する
            if (0 < buffer.Length)
              continue;
          }
        }
      }
      catch (Exception ex) {
        Trace.Log(ex);
        Trace.Error("parser error. buffered message was: {0}", buffer);

        // TODO: parser errorの場合の考慮
        buffer = string.Empty; // 捨てる
      }

      return responses.ToArray();
    }

    private ImapProtocolResponse TryParse(string buffer, out string unparsed)
    {
      if (string.Empty.Equals(buffer)) {
        unparsed = string.Empty;
        return null;
      }

      var crlf = buffer.IndexOf("\r\n");
      string firstLine;
      string trailingLines;

      if (crlf == -1) {
        firstLine = buffer;
        trailingLines = string.Empty;
      }
      else {
        firstLine = buffer.Substring(0, crlf + 2);
        trailingLines = buffer.Substring(crlf + 2);
      }

      if (firstLine.StartsWith("+ ")) {
        // continue-req    = "+" SP (resp-text / base64) CRLF
        if (crlf == -1) {
          unparsed = buffer;
          return null;
        }
        else {
          unparsed = trailingLines;
          return new ImapCommandContinuationRequest(firstLine.Substring(2, firstLine.Length - 4), // remove "+", SP, CRLF
                                                    firstLine);
        }
      }

      unparsed = buffer;

      // ("*" / tag) SP
      var tagSep = firstLine.IndexOf(' ');

      if (tagSep == -1)
        return null;

      var tag = firstLine.Substring(0, tagSep);

      // ("OK" / "BAD" / "NO" / "BYE" / "PREAUTH" / text) SP
      var respCondSep = firstLine.IndexOf(' ', tagSep + 1);

      if (respCondSep == -1)
        return null;

      var respCond = firstLine.Substring(tagSep + 1, respCondSep - tagSep - 1);

      var status = ImapProtocolStatus.None;

      switch (respCond.ToUpper()) {
        case "OK":
          status = ImapProtocolStatus.OK;
          break;
        case "BAD":
          status = ImapProtocolStatus.Bad;
          break;
        case "NO":
          status = ImapProtocolStatus.No;
          break;
        case "BYE":
          status = ImapProtocolStatus.Bye;
          break;
        case "PREAUTH":
          status = ImapProtocolStatus.PreAuth;
          break;
        default:
          status = ImapProtocolStatus.None;
          break;
      }

      if (status != ImapProtocolStatus.None ||
          firstLine.StartsWith("* [")) {
        // resp-cond-auth / resp-cond-state / resp-cond-bye
        if (crlf == -1)
          return null;

        string respText;

        if (status == ImapProtocolStatus.None)
          respText = firstLine.Substring(tagSep + 1);
        else
          respText = firstLine.Substring(respCondSep + 1);

        ImapProtocolResponseCode responseCode;
        string text;

        ParseRespText(respText, out responseCode, out text);

        unparsed = trailingLines;

        return new ImapStatusResponse(tag,
                                      status,
                                      responseCode,
                                      text,
                                      firstLine);
      }
      else {
        // mailbox-data / message-data / capability-data
        if (!trailingLines.EndsWith("\r\n"))
          return null;

        var data = firstLine + trailingLines;
        var dataReader = new StringReader(data.Substring(tagSep + 1));

        ImapData[] parsed = null;

        if (ParseData(dataReader, out parsed)) {
          unparsed = dataReader.ReadToEnd();
          return new ImapDataResponse(parsed,
                                      data.Substring(0, data.Length - unparsed.Length));
        }
        else {
          unparsed = firstLine + trailingLines;
          return null;
        }
      }
    }

    private void ParseRespText(string respText, out ImapProtocolResponseCode respCode, out string text)
    {
      // resp-text       = ["[" resp-text-code "]" SP] text

      respCode = null;
      text = string.Empty;

      if (respText.StartsWith("[")) {
        // exists resp-text-code
        var respTextCodeEnd = respText.IndexOf("]");
        var respTextCode    = respText.Substring(1, respTextCodeEnd - 1);

        if (respTextCodeEnd < respText.Length) {
          text = respText.Substring(respTextCodeEnd + 1);
          text = text.Trim(); // remove SP, CRLF
        }
        else {
          text = string.Empty;
        }

        var codeSep = respTextCode.IndexOf(" ");

        if (codeSep == -1) {
          // "READ-ONLY" / "READ-WRITE" / etc...
          respCode = new ImapProtocolResponseCode(respTextCode, new ImapData[] {});
        }
        else {
          // "UIDNEXT" SP nz-number / "UIDVALIDITY" SP nz-number / etc...
          var code        = respTextCode.Substring(0, codeSep);
          var dataString  = respTextCode.Substring(codeSep + 1);

          ImapData[] data = null;

          if (!ParseData(new StringReader(dataString), out data))
            return;

          respCode = new ImapProtocolResponseCode(code, data);
        }
      }
      else {
        // no resp-text-code
        respCode = null;
        text = respText.Substring(0, respText.Length - 2); // remove CRLF
      }
    }

    private bool ParseData(StringReader dataReader, out ImapData[] data)
    {
      data = null;

      var parsed = new List<ImapData>();

      for (;;) {
        var c = dataReader.Peek();

        if (c == -1)
          break;

        var ch = (char)c;

        if (ch == ' ' || ch == '\r') { // SP, CR
          dataReader.Read();
          continue;
        }

        if (ch == '\n') { // LF
          dataReader.Read();
          break;
        }

        ImapData d;
        var builder = new StringBuilder();

        if (ch == '"') {
          if (ParseQuotedString(dataReader, builder, false))
            d = ImapData.CreateTextData(builder.ToString());
          else
            return false;
        }
        else if (ch == '(') {
          if (ParseParenthesizedData(dataReader, builder)) {
            ImapData[] list = null;

            if (!ParseData(new StringReader(builder.ToString()), out list))
              return false;

            d = ImapData.CreateListData(list);
          }
          else {
            return false;
          }
        }
        else if (ch == '{') {
          if (ParseLiteralString(dataReader, builder, false))
            d = ImapData.CreateTextData(builder.ToString());
          else
            return false;
        }
        else {
          if (ParseNonQuotedText(dataReader, builder)) {
            var text = builder.ToString();

            if (text == "NIL")
              d = ImapData.CreateNilData();
            else
              d = ImapData.CreateTextData(text);
          }
          else {
            return false;
          }
        }

        parsed.Add(d);
      }

      data = parsed.ToArray();

      return true;
    }

    private bool ParseQuotedString(StringReader reader, StringBuilder builder, bool keepDelimiter)
    {
      // quoted          = DQUOTE *QUOTED-CHAR DQUOTE
      // QUOTED-CHAR     = <any TEXT-CHAR except quoted-specials> / 
      //                   "\" quoted-specials
      // TEXT-CHAR       = <any CHAR except CR and LF>
      // CHAR            = <any 7-bit US-ASCII character except NUL,
      //                   0x01 - 0x7f>
      // quoted-specials = DQUOTE / "\"
      // date-time       = DQUOTE date-day-fixed "-" date-month "-" date-year

      var c = reader.Read(); // DQUOTE

      if (keepDelimiter)
        builder.Append((char)c);

      for (;;) {
        c = reader.Read();

        if (c == -1)
          return true; // suchess

        var ch = (char)c;

        if (ch == '\r' || ch == '\n' || ch == '"') { // CR, LF, DQUOTE
          if (keepDelimiter)
            builder.Append(ch);

          return true; // suchess
        }
        else { // QUOTED-CHAR
          builder.Append(ch);
        }
      }
    }

    private bool ParseLiteralString(StringReader reader, StringBuilder builder, bool keepDelimiter)
    {
      // literal         = "{" number "}" CRLF *CHAR8
      //                ; Number represents the number of CHAR8s
      // CHAR8           = %x01-ff ; any OCTET except NUL, %x00

      var c = reader.Read(); // "{"

      var octetString = string.Empty;

      for (;;) {
        c = reader.Read();

        if (0x30 <= c && c <= 0x39)
          octetString += (char)c; // number
        else if ((char)c == '}')
          break; // "}"
        else
          return false; // abort
      }

      if (keepDelimiter) {
        builder.Append('{');
        builder.Append(octetString);
        builder.Append('}');
      }

      var octet = int.Parse(octetString);

      // CRLF
      if (0x0d != reader.Read())
        return false; // abort

      if (keepDelimiter)
        builder.Append('\r');

      if (0x0a != reader.Read())
        return false; // abort

      if (keepDelimiter)
        builder.Append('\n');

      // *CHAR8
      for (var o = 0; ; o++) {
        if (o == octet)
          return true; // success

        c = reader.Read();

        if (c == -1)
          return false; // abort

        builder.Append((char)c);
      }
    }

    private bool ParseParenthesizedData(StringReader reader, StringBuilder builder)
    {
      var nested = 0;
      var c = reader.Read(); // "("

      for (;;) {
        c = reader.Peek();

        if (c == -1)
          return false; // abort

        if ((char)c == '"') {
          // quoted
          if (ParseQuotedString(reader, builder, true))
            continue;
          else
            return false;
        }
        else if ((char)c == '{') {
          // literal
          if (ParseLiteralString(reader, builder, true))
            continue;
          else
            return false;
        }

        var ch = (char)reader.Read();

        if (ch == '(') {
          nested++;
        }
        else if (ch == ')') {
          nested--;
          if (nested == -1) { // not nested
            return true; // suchess
          }
        }

        builder.Append((char)c);
      }
    }

    private bool ParseNonQuotedText(StringReader reader, StringBuilder builder)
    {
      // atom            = 1*ATOM-CHAR
      // ATOM-CHAR       = <any CHAR except atom-specials>
      // atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
      //                   quoted-specials / resp-specials
      // number          = 1*DIGIT
      // list-wildcards  = "%" / "*"
      // quoted-specials = DQUOTE / "\"
      // resp-specials   = "]"
      for (;;) {
        var c = reader.Peek();

        if (c == -1)
          break;
        else if (c == 0x0d)
          break; // CR
        else if (c == 0x0a)
          break; // LF

        char ch = (char)c;

        if (ch == ' ') // SP
          break;
        else
          builder.Append((char)reader.Read());

        /*
        if (0x30 <= c && c <= 0x39) {
          // '0' 0x30 to '9' 0x39
          builder.Append((char)reader.Read());
          continue;
        }
        else if (0x41 <= c && c <= 0x5a) {
          // 'A' 0x41 to 'Z' 0x5a
          builder.Append((char)reader.Read());
          continue;
        }
        else if (0x61 <= c && c <= 0x7a) {
          // 'a' 0x61 to 'z' 0x7a
          builder.Append((char)reader.Read());
          continue;
        }

        char ch = (char)c;

        if (ch == ' ') // SP
          break;
        else if (ch == '(' || ch == ')' || ch == '{' ||
                 ch == '%' || ch == '*' || ch == '\\' ||
                 ch == '"' || ch == ']')
          break;
        else if (Char.IsControl(ch))
          break;
        else
          builder.Append((char)reader.Read());
          */
      }

      return true; // suchess
    }

    private static readonly Encoding latin1Encoding = Encoding.GetEncoding("ISO-8859-1");
    private string buffer = string.Empty;
  }
}