using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4.Protocol {
  // format converting methods for ImapStatusResponse
  internal static class ImapStatusResponseConverter {
    /*
     * 7.1. Server Responses - Status Responses
     */

    //      ALERT
    //         The human-readable text contains a special alert that MUST be
    //         presented to the user in a fashion that calls the user's
    //         attention to the message.
    public static bool FromAlert(ImapStatusResponse status, out string message)
    {
      message = null;

      if (!IsRequiredStatusResponse(status, "ALERT"))
        return false;

      message = status.Text;

      return true;
    }

    //      BADCHARSET
    //         Optionally followed by a parenthesized list of charsets.  A
    //         SEARCH failed because the given charset is not supported by
    //         this implementation.  If the optional list of charsets is
    //         given, this lists the charsets that are supported by this
    //         implementation.
    public static bool FromBadCharset(ImapStatusResponse status, out string[] supportedCharsets)
    {
      supportedCharsets = null;

      if (!IsRequiredStatusResponse(status, "BADCHARSET"))
        return false;

      var list = ImapDataConverter.ToTextList(status.ResponseCode.Data);

      if (list == null) {
        return false;
      }
      else {
        supportedCharsets = list;
        return true;
      }
    }

    //      CAPABILITY
    //         Followed by a list of capabilities.  This can appear in the
    //         initial OK or PREAUTH response to transmit an initial
    //         capabilities list.  This makes it unnecessary for a client to
    //         send a separate CAPABILITY command if it recognizes this
    //         response.
    public static bool FromCapability(ImapStatusResponse status, out ImapCapability capabilities)
    {
      capabilities = null;

      if (!IsRequiredStatusResponse(status, "CAPABILITY"))
        return false;

      // insert status text 'CAPABILITY' to response data list
      var list = new List<ImapData>(status.ResponseCode.Data);

      list.Insert(0, ImapData.CreateTextData("CAPABILITY"));

      return ImapDataConverter.ToCapabilityData(list.ToArray(), out capabilities);
    }

    //      PARSE
    //         The human-readable text represents an error in parsing the
    //         [RFC-2822] header or [MIME-IMB] headers of a message in the
    //         mailbox.
    public static bool FromParse(ImapStatusResponse status, out string message)
    {
      message = null;

      if (!IsRequiredStatusResponse(status, "PARSE"))
        return false;

      message = status.Text;

      return true;
    }

    //      PERMANENTFLAGS
    //         Followed by a parenthesized list of flags, indicates which of
    //         the known flags the client can change permanently.  Any flags
    //         that are in the FLAGS untagged response, but not the
    //         PERMANENTFLAGS list, can not be set permanently.  If the client
    //         attempts to STORE a flag that is not in the PERMANENTFLAGS
    //         list, the server will either ignore the change or store the
    //         state change for the remainder of the current session only.
    //         The PERMANENTFLAGS list can also include the special flag \*,
    //         which indicates that it is possible to create new keywords by
    //         attempting to store those flags in the mailbox.
    public static bool FromPermanentFlags(ImapStatusResponse status, out ImapMessageFlags flags, out string[] extendedFlags)
    {
      flags = ImapMessageFlags.None;
      extendedFlags = null;

      if (!IsRequiredStatusResponse(status, "PERMANENTFLAGS"))
        return false;

      return ImapDataConverter.ToFlag(status.ResponseCode.Data, out flags, out extendedFlags);
    }

    //      READ-ONLY
    //         The mailbox is selected read-only, or its access while selected
    //         has changed from read-write to read-only.
    public static bool FromReadOnly(ImapStatusResponse status)
    {
      return IsRequiredStatusResponse(status, "READ-ONLY");
    }

    //      READ-WRITE
    //         The mailbox is selected read-write, or its access while
    //         selected has changed from read-only to read-write.
    public static bool FromReadWrite(ImapStatusResponse status)
    {
      return IsRequiredStatusResponse(status, "READ-WRITE");
    }

    //      TRYCREATE
    //         An APPEND or COPY attempt is failing because the target mailbox
    //         does not exist (as opposed to some other reason).  This is a
    //         hint to the client that the operation can succeed if the
    //         mailbox is first created by the CREATE command.
    public static bool FromTryCreate(ImapStatusResponse status)
    {
      return IsRequiredStatusResponse(status, "TRYCREATE");
    }

    //      UIDNEXT
    //         Followed by a decimal number, indicates the next unique
    //         identifier value.  Refer to section 2.3.1.1 for more
    //         information.
    public static bool FromUidNext(ImapStatusResponse status, out int uidNext)
    {
      uidNext = 0;

      if (!IsRequiredStatusResponse(status, "UIDNEXT"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      var val = ImapDataConverter.ToNumber(status.ResponseCode.Data[0]);

      if (val == null) {
        return false;
      }
      else {
        uidNext = (int)val;
        return true;
      }
    }

    //      UIDVALIDITY
    //         Followed by a decimal number, indicates the unique identifier
    //         validity value.  Refer to section 2.3.1.1 for more information.
    public static bool FromUidValidity(ImapStatusResponse status, out int uidValidity)
    {
      uidValidity = 0;

      if (!IsRequiredStatusResponse(status, "UIDVALIDITY"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      var val = ImapDataConverter.ToNumber(status.ResponseCode.Data[0]);

      if (val == null) {
        return false;
      }
      else {
        uidValidity = (int)val;
        return true;
      }
    }

    //      UNSEEN
    //         Followed by a decimal number, indicates the number of the first
    //         message without the \Seen flag set.
    public static bool FromUnseen(ImapStatusResponse status, out int unseen)
    {
      unseen = 0;

      if (!IsRequiredStatusResponse(status, "UNSEEN"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      var val = ImapDataConverter.ToNumber(status.ResponseCode.Data[0]);
    
      if (val == null) {
        return false;
      }
      else {
        unseen = (int)val;
        return true;
      }
    }

    /*
     * required/allowed command responses
     */
    public static bool FromGreetingResponse(ImapStatusResponse status, out ImapCapability capabilities)
    {
      capabilities = null;

      if (status.Status != ImapProtocolStatus.OK)
        return false;

      // 'This can appear in the initial OK or PREAUTH response'
      return FromCapability(status, out capabilities);
    }

    public static bool FromSearchResponse(ImapStatusResponse status, out string[] supportedCharset)
    {
      supportedCharset = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // 'A SEARCH failed because the given charset is not
      //  supported by this implementation'
      return FromBadCharset(status, out supportedCharset);
    }

    // 6.3.1. SELECT Command
    // Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
    //             REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
    //             UIDNEXT, UIDVALIDITY
    public static bool FromSelectExamineResponse(ImapStatusResponse status, out bool? readOnly, out int? unseen, out ImapMessageFlags? permanentFlags, out string[] extendedPermanentFlags, out int? uidNext, out int? uidValidity)
    {
      readOnly                = null;
      unseen                  = null;
      permanentFlags          = null;
      extendedPermanentFlags  = null;
      uidNext                 = null;
      uidValidity             = null;

      if (status.Status != ImapProtocolStatus.OK)
        return false;

      // readonly, readwrite
      // READ-ONLY
      if (FromReadOnly(status)) {
        readOnly = true;
        return true;
      }

      // READ-WRITE
      if (FromReadWrite(status)) {
        readOnly = false;
        return true;
      }

      // PERMANENTFLAGS
      ImapMessageFlags respFlags;
      string[] respExtendedFlags;

      if (FromPermanentFlags(status, out respFlags, out respExtendedFlags)) {
        permanentFlags = respFlags;
        extendedPermanentFlags = respExtendedFlags;
        return true;
      }

      // UIDNEXT
      int respUidNext;

      if (FromUidNext(status, out respUidNext)) {
        uidNext = respUidNext;
        return true;
      }

      // UIDVALIDITY
      int respUidValidity;

      if (FromUidNext(status, out respUidValidity)) {
        uidValidity = respUidValidity;
        return true;
      }

      // UNSEEN
      int respUnseen;

      if (FromUnseen(status, out respUnseen)) {
        unseen = respUnseen;
        return true;
      }

      return false;
    }

    public static bool FromAppendResponse(ImapStatusResponse status)
    {
      /*
      If the destination mailbox does not exist, a server MUST return an
      error, and MUST NOT automatically create the mailbox.  Unless it
      is certain that the destination mailbox can not be created, the
      server MUST send the response code "[TRYCREATE]" as the prefix of
      the text of the tagged NO response.  This gives a hint to the
      client that it can attempt a CREATE command and retry the APPEND
      if the CREATE is successful.
      */
      if (status.Status != ImapProtocolStatus.No)
        return false;

      return FromTryCreate(status);
    }

    public static bool FromCopyResponse(ImapStatusResponse status)
    {
      /*
      If the destination mailbox does not exist, a server SHOULD return
      an error.  It SHOULD NOT automatically create the mailbox.  Unless
      it is certain that the destination mailbox can not be created, the
      server MUST send the response code "[TRYCREATE]" as the prefix of
      the text of the tagged NO response.  This gives a hint to the
      client that it can attempt a CREATE command and retry the COPY if
      the CREATE is successful.
      */
      if (status.Status != ImapProtocolStatus.No)
        return false;

      return FromTryCreate(status);
    }

    /*
     * utility methods
     */

    private static bool IsRequiredStatusResponse(ImapStatusResponse status, string code)
    {
      if (status.ResponseCode == null)
        return false;

      if (!status.ResponseCode.Code.Equals(code, StringComparison.InvariantCultureIgnoreCase))
        return false;

      return true;
    }
  }
}