using System;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal class IdleTransaction : TransactionBase {

    // ProcessAlreadyIdling
    //   x Conclude
    // ProcessMailboxNotFound
    //   x Conclude
    // ProcessIdleIncapable
    //   x Conclude
    // ProcessIdle
    //   ->WaitForIdle
    //   x Conclude
    // WaitForIdle
    //   x Conclude

    public IdleTransaction(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if ((Connection.ServerInfo.Capability.KnownCapabilities & ImapKnownCapabilities.Idle) != 0) {
        if (Connection.IsIdling) {
          return ProcessAlreadyIdling;
        }
        else {
          if (Request.Parameters.ContainsKey("Mailbox")) {
            idlingMailbox = ImapMailboxManager.GetExistMailbox(Connection.ServerInfo, Request.Parameters["Mailbox"]);

            if (idlingMailbox == null)
              return ProcessMailboxNotFound;
          }
          else {
            idlingMailbox = null;
          }

          return ProcessIdle;
        }
      }
      else {
        return ProcessIdleIncapable;
      }
    }

    private void ProcessAlreadyIdling()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.Done, "already idling"));
    }

    private void ProcessMailboxNotFound()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.NotFound, string.Format("mailbox '{0}' was not found", Request.Parameters["Mailbox"])));
    }

    private void ProcessIdleIncapable()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.Incapable, "IDLE incapable"));
    }

    private void ProcessIdle()
    {
      lastTag = Send("idle");

      Transit(WaitForIdle);
    }

    private void WaitForIdle()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapCommandContinuationRequest) {
        Connection.IsIdling = true;
      }
      else if (response is ImapDataResponse) {
        if (idlingMailbox != null) {
          int? exists;
          int? expunge;

          if (ImapDataResponseConverter.FromIdleCommandResponse(response as ImapDataResponse,
                                                                out exists,
                                                                out expunge)) {
            if (exists != null) idlingMailbox.ExistsMessage = (int)exists;
            // TODO: if (expunge != null)

            Trace.Dump(idlingMailbox);
          }
        }

        // callback
        if (Connection.IdlingCallback != null)
          Connection.IdlingCallback.BeginInvoke(idlingMailbox, null, null);
      }
      else if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.Tag != lastTag)
          return;

        Connection.IsIdling = false;

        if (status.Status == ImapProtocolStatus.OK) {
          Conclude(new ImapResponse(ImapTransactionStatus.Ok, "idling"));
        }
        else {
          ConcludeError(status);
        }
      }
    }

    private string lastTag = string.Empty;
    private ImapMailbox idlingMailbox = null;
  }
}
