using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net.Security;
using System.IO;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Web;

using Smdn.Protocols.Pop3.Transaction;

namespace Smdn.Protocols.Pop3.Session {
  // Layers:
  //   Client.PopClient
  //   Client.PopClientAgent
  // * Session.PopSessionAgent
  //   Session.PopSession
  //   Transaction.PopTransactionAgent
  //   Transaction.PopTransaction/Transaction.PopRequest/Transaction.PopResponse
  //   Protocol.PopConnection/Protocol.PopProtocolCommand/Protocol.PopProtocolResponse
  //   (POP3)

  public class PopSessionAgent : IDisposable {
    public X509Certificate2Collection CertCollection {
      get { return certCollection; }
    }

    public RemoteCertificateValidationCallback ServerCertificateValidationCallback {
      get { return serverCertificateValidationCallback; }
      set { serverCertificateValidationCallback = value; }
    }

    public PopCreateAuthenticatedStreamCallback CreateAuthenticatedStreamCallback {
      get { return createAuthenticatedStreamCallback; }
      set { createAuthenticatedStreamCallback = value; }
    }

    /*
     * construction/destruction
     */

    public PopSessionAgent()
    {
    }

    public void Initialize()
    {
      if (agent != null)
        throw new InvalidOperationException("already initialized");

      Trace.Info("PopSessionAgent initializing");

      agent = new PopTransactionAgent();
      agent.Initialize();

      Trace.Info("PopSessionAgent initialized");
    }

    ~PopSessionAgent()
    {
      Dispose();
      GC.SuppressFinalize(this);
    }

    public void Dispose()
    {
      if (disposed)
        return;

      Trace.Info("PopSessionAgent disposing");

      if (agent != null) {
        List<PopSession> existSessions;

        lock (sessions) {
          existSessions = new List<PopSession>(sessions.Values);
        }

        foreach (var session in existSessions) {
          if (session.State == PopSessionState.Transaction)
            Quit(session);
          if (session.State != PopSessionState.NotConnected)
            Disconnect(session);
        }

        lock (sessions) {
          sessions.Clear();
        }

        agent.Dispose();
        agent = null;
      }

      Trace.Info("PopSessionAgent disposed");

      disposed = true;
    }

    /*
     * method for connect
     */

    /// <summary>connects to POP server</summary>
    public bool Connect(PopUri uri, out PopSession session)
    {
      session = GetExistSession(uri);

      if (session != null)
        return true;

      session = PopSession.Start(uri, CreateAuthenticatedStream);

      if (session == null)
        return false;

      TraceVerbose(session, "waiting for initial server greeting response");

      PopResponse response;

      if (RequestSync(session, PopTransactionMethod.Greeting, out response)) {
        lock (sessions) {
          sessions.Add(uri.UserAtHost, session);
        }

        TransitStateTo(session, PopSessionState.Authorization);

        return true;
      }
      else {
        session = null;

        return false;
      }
    }

    /*
     * requesting methods for any state
     */

    /// <summary>sends QUIT command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Quit(PopSession session)
    {
      if (session.State == PopSessionState.NotConnected) {
        TraceVerbose(session, "already quitted or disconnected");
        return true;
      }

      PopResponse response;

      RequestSync(session, PopTransactionMethod.Quit, out response);

      if (response.Status == PopTransactionStatus.Ok) {
        Disconnect(session);

        return true;
      }
      else {
        return false;
      }
    }

    /// <summary>sends CAPA command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Capa(PopSession session, out PopCapability capability)
    {
      RejectNonConnectedState(session);

      capability = null;

      PopResponse response;

      if (RequestSync(session, PopTransactionMethod.Capa, out response)) {
        var resp = (response as PopCapabilityResponse);

        capability = resp.Capability;

        return true;
      }

      return false;
    }

    /*
     * requesting methods for authrozation state
     */

    /// <summary>sends USER and PASS command</summary>
    /// <remarks>valid in authrozation state</remarks>
    public bool UserPass(PopSession session, string password)
    {
      var request = CheckStateAndCreateLoginRequest(session, password, PopTransactionMethod.UserPass);

      if (request == null)
        return true;

      if (RequestSync(session, request)) {
        TransitStateTo(session, PopSessionState.Transaction);
        return true;
      }

      return false;
    }

    /// <summary>sends APOP command</summary>
    /// <remarks>valid in authrozation state</remarks>
    public bool Apop(PopSession session, string password)
    {
      var request = CheckStateAndCreateLoginRequest(session, password, PopTransactionMethod.Apop);

      if (request == null)
        return true;

      if (RequestSync(session, request)) {
        TransitStateTo(session, PopSessionState.Transaction);
        return true;
      }

      return false;
    }

    /// <summary>sends AUTH command</summary>
    /// <remarks>valid in authrozation state</remarks>
    public bool Auth(PopSession session, string password)
    {
      var request = CheckStateAndCreateLoginRequest(session, password, PopTransactionMethod.Auth);

      if (request == null)
        return true;

      switch (session.HostUri.Authentication) {
        case PopAuthenticationType.None:
        case PopAuthenticationType.Any:
          request.Parameters["Auth-Type"] = string.Empty;
          break;
        default:
          request.Parameters["Auth-Type"] = PopConverter.ToString(session.HostUri.Authentication);
          break;
      }

      if (RequestSync(session, request)) {
        TransitStateTo(session, PopSessionState.Transaction);
        return true;
      }

      return false;
    }

    private PopRequest CheckStateAndCreateLoginRequest(PopSession session, string password, PopTransactionMethod method)
    {
      RejectNonConnectedState(session);

      if (session.State == PopSessionState.Transaction) {
        TraceVerbose(session, "already authenticated");
        return null;
      }

      var request = new PopRequest(method, session.Timeout);

      request.Parameters["Mailbox"] = session.HostUri.UserName;
      request.Parameters["Password"] = password;

      return request;
    }

    /*
     * requesting methods for transaction state
     */

    /// <summary>sends STAT command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Stat(PopSession session, out PopDropList dropList)
    {
      RejectNonTransactionState(session);

      dropList = null;

      PopResponse response;

      if (RequestSync(session, PopTransactionMethod.Stat, out response)) {
        var resp = (response as PopDropListResponse);

        dropList = resp.DropList;

        return true;
      }

      return false;
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool List(PopSession session, int messageNumber, out PopScanList scanList)
    {
      RejectNonTransactionState(session);

      scanList = null;

      var request = new PopRequest(PopTransactionMethod.List, session.Timeout);

      request.Parameters["Message-Number"] = messageNumber.ToString();

      PopResponse response;

      if (RequestSync(session, request, out response)) {
        var resp = (response as PopScanListResponse);

        scanList = resp.ScanLists[0];

        return true;
      }

      return false;
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool List(PopSession session, out PopScanList[] scanLists)
    {
      RejectNonTransactionState(session);

      scanLists = null;

      PopResponse response;

      if (RequestSync(session, PopTransactionMethod.List, out response)) {
        var resp = (response as PopScanListResponse);

        scanLists = resp.ScanLists;

        return true;
      }

      return false;
    }

    /// <summary>sends RETR command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Retr(PopSession session, PopScanList scanList, out PopMessage message)
    {
      return Retr(session, scanList.MessageNumber, out message);
    }

    /// <summary>sends RETR command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Retr(PopSession session, int messageNumber, out PopMessage message)
    {
      RejectNonTransactionState(session);

      message = null;

      var request = new PopRequest(PopTransactionMethod.Retr, session.Timeout);

      request.Parameters["Message-Number"] = messageNumber.ToString();

      PopResponse response;

      if (RequestSync(session, request, out response)) {
        var resp = (response as PopMessageResponse);

        message = resp.Message;

        return true;
      }

      return false;
    }

    /// <summary>sends DELE command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Dele(PopSession session, PopMessage message)
    {
      return Dele(session, message.Number);
    }

    /// <summary>sends DELE command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Dele(PopSession session, PopScanList scanList)
    {
      return Dele(session, scanList.MessageNumber);
    }

    /// <summary>sends DELE command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Dele(PopSession session, int messageNumber)
    {
      RejectNonTransactionState(session);

      var request = new PopRequest(PopTransactionMethod.Dele, session.Timeout);

      request.Parameters["Message-Number"] = messageNumber.ToString();

      return RequestSync(session, request);
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool NoOp(PopSession session)
    {
      RejectNonTransactionState(session);

      return RequestSync(session, PopTransactionMethod.NoOp);
    }

    /// <summary>sends RSET command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Rset(PopSession session)
    {
      RejectNonTransactionState(session);

      return RequestSync(session, PopTransactionMethod.Rset);
    }

    /// <summary>sends TOP command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Top(PopSession session, PopScanList scanList, int lines, out PopMessage message)
    {
      return Top(session, scanList.MessageNumber, lines, out message);
    }

    /// <summary>sends TOP command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Top(PopSession session, int messageNumber, int lines, out PopMessage message)
    {
      RejectNonTransactionState(session);

      message = null;

      var request = new PopRequest(PopTransactionMethod.Top, session.Timeout);

      request.Parameters["Message-Number"] = messageNumber.ToString();
      request.Parameters["Lines"] = lines.ToString();

      PopResponse response;

      if (RequestSync(session, request, out response)) {
        var resp = (response as PopMessageResponse);

        message = resp.Message;

        return true;
      }

      return false;
    }

    /// <summary>sends UIDL command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Uidl(PopSession session, PopMessage message, out PopUniqueIdList uidList)
    {
      return Uidl(session, message.Number, out uidList);
    }

    /// <summary>sends UIDL command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Uidl(PopSession session, PopScanList scanList, out PopUniqueIdList uidList)
    {
      return Uidl(session, scanList.MessageNumber, out uidList);
    }

    /// <summary>sends UIDL command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Uidl(PopSession session, int messageNumber, out PopUniqueIdList uidList)
    {
      RejectNonTransactionState(session);

      uidList = null;

      var request = new PopRequest(PopTransactionMethod.Uidl, session.Timeout);

      request.Parameters["Message-Number"] = messageNumber.ToString();

      PopResponse response;

      if (RequestSync(session, request, out response)) {
        var resp = (response as PopUniqueIdListResponse);

        uidList = resp.UniqueIdLists[0];

        return true;
      }

      return false;
    }

    /// <summary>sends UIDL command</summary>
    /// <remarks>valid in transaction state</remarks>
    public bool Uidl(PopSession session, out PopUniqueIdList[] uidLists)
    {
      RejectNonTransactionState(session);

      uidLists = null;

      PopResponse response;

      if (RequestSync(session, PopTransactionMethod.Uidl, out response)) {
        var resp = (response as PopUniqueIdListResponse);

        uidLists = resp.UniqueIdLists;

        return true;
      }

      return false;
    }

    /*
     * internal requesting methods
     */

    private PopTransaction RequestAsync(PopSession session, PopRequest request)
    {
      RejectDisposed();

      return agent.Request(session.Connection, request);
    }

    private bool RequestSync(PopSession session, PopTransactionMethod method)
    {
      PopResponse response;

      return RequestSync(session, method, out response);
    }

    private bool RequestSync(PopSession session, PopTransactionMethod method, out PopResponse response)
    {
      return RequestSync(session, new PopRequest(method, session.Timeout), out response);
    }

    private bool RequestSync(PopSession session, PopRequest request)
    {
      PopResponse response;

      return RequestSync(session, request, out response);
    }

    private bool RequestSync(PopSession session, PopRequest request, out PopResponse response)
    {
      RejectDisposed();

      response = null;

      for (var retry = 0; retry < 3; retry++) {
        var transaction = agent.Request(session.Connection, request);

        while (transaction.Response == null) {
          Thread.Sleep(20);
        }

        response = transaction.Response;

        if (response.Status == PopTransactionStatus.TransactionProgress) {
          // retry
          Thread.Sleep(250);
          continue;
        }
        else {
          break;
        }
      }

      // true for 2xx, false for 3xx/4xx/5xx
      return ((int)response.Status < 300);
    }

    /*
     * internal methods
     */

    internal PopSession GetExistSession(PopUri uri)
    {
      lock (sessions) {
        if (sessions.ContainsKey(uri.UserAtHost))
          return sessions[uri.UserAtHost];
      }

      return null;
    }

    /*
     * methods for checking status
     */

    private static void RejectNonConnectedState(PopSession session)
    {
      if (session == null)
        throw new NullReferenceException("session is null");

      if (session.State == PopSessionState.NotConnected)
        throw new InvalidOperationException("session has been disconnected");
    }

    private static void RejectNonAuthorizationState(PopSession session)
    {
      RejectNonConnectedState(session);

      if (session.State != PopSessionState.Authorization)
        throw new InvalidOperationException("session is not in authorization state");
    }

    private static void RejectNonTransactionState(PopSession session)
    {
      RejectNonConnectedState(session);

      if (session.State != PopSessionState.Transaction)
        throw new InvalidOperationException("session is not in transaction state");
    }

    /*
     * private methods
     */

    private void Disconnect(PopSession session)
    {
      lock (sessions) {
        if (sessions.ContainsKey(session.HostUri.UserAtHost)) {
          sessions.Remove(session.HostUri.UserAtHost);
        }
      }

      session.Close();

      TraceVerbose(session, "now in non-connected state");
    }

    private void TransitStateTo(PopSession session, PopSessionState state)
    {
      switch (state) {
        case PopSessionState.Authorization:
          TraceVerbose(session, "now in authorization state");
          break;

        case PopSessionState.Transaction:
          TraceVerbose(session, "now in transaction state");
          break;
      }

      session.State = state;
    }

    private Stream CreateAuthenticatedStream(PopUri uri, Stream baseStream)
    {
      if (createAuthenticatedStreamCallback != null) {
        return createAuthenticatedStreamCallback(uri, baseStream);
      }
      else {
        const int popsPort = 995;

        if (uri.Port == popsPort) {
          SslStream sslStream = null;

          if (serverCertificateValidationCallback == null)
            sslStream = new SslStream(baseStream, false);
          else
            sslStream = new SslStream(baseStream, false, serverCertificateValidationCallback);

          sslStream.AuthenticateAsClient(uri.Host, certCollection, SslProtocols.Default, true);

          Trace.Verbose("[PopSession] Created SSL stream for {0}", uri);
          Trace.Verbose("[PopSession]   IsAuthenticated: {0}", sslStream.IsAuthenticated);
          Trace.Verbose("[PopSession]   IsMutuallyAuthenticated: {0}", sslStream.IsMutuallyAuthenticated);
          Trace.Verbose("[PopSession]   IsEncrypted: {0}", sslStream.IsEncrypted);
          Trace.Verbose("[PopSession]   IsSigned: {0}", sslStream.IsSigned);

          return sslStream;
        }
        else {
          return baseStream;
        }
      }
    }

    [Conditional("TRACE")]
    private void TraceVerbose(PopSession session, string format, params object[] arguments)
    {
      Trace.Verbose("CID:{0} （≧▽≦）[PopSession] {1}", session.Id, string.Format(format, arguments));
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private PopTransactionAgent agent = null;
    private readonly Dictionary<string, PopSession> sessions = new Dictionary<string, PopSession>();
    private readonly X509Certificate2Collection certCollection = new X509Certificate2Collection();
    private RemoteCertificateValidationCallback serverCertificateValidationCallback = null;
    private PopCreateAuthenticatedStreamCallback createAuthenticatedStreamCallback = null;
  }
}