using System;
using System.Collections.Generic;

using Smdn.Protocols.Pop3.Protocol;

namespace Smdn.Protocols.Pop3.Transaction.Transactions {
  internal class UidlTransaction : TransactionBase {

    // ProcessUidlIncapable
    //   x Conclude
    // ProcessUidl
    //   ->WaitForUidl
    // WaitForUidl
    //   x Conclude
    // RetrieveUidl
    //   x Conclude

    public UidlTransaction(PopConnection connection, PopRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if ((Connection.ServerInfo.Capability.KnownCapabilities & PopKnownCapabilities.Capabilities) != 0 &&
          (Connection.ServerInfo.Capability.KnownCapabilities & PopKnownCapabilities.Uidl) == 0)
        return ProcessUidlIncapable;

      listMultiple = !Request.Parameters.ContainsKey("Message-Number");

      return ProcessUidl;
    }

    private void ProcessUidlIncapable()
    {
      Conclude(new PopResponse(PopTransactionStatus.Incapable, "UIDL incapable"));
    }

    private void ProcessUidl()
    {
      if (listMultiple)
        Send("UIDL");
      else
        Send("UIDL", Request.Parameters["Message-Number"]);

      Transit(WaitForUidl);
    }

    private void WaitForUidl()
    {
      PopProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is PopStatusResponse) {
        var status = response as PopStatusResponse;

        if (status.Status == PopProtocolStatus.Positive) {
          if (listMultiple) {
            Transit(RetrieveUidl);
          }
          else {
            PopUniqueIdList uidList;
            if (RetrieveUniqueIdList(status, out uidList))
              Conclude(new PopUniqueIdListResponse(PopTransactionStatus.Ok, new[] {uidList}, "etrieved a unique-id list"));
            else
              Conclude(new PopResponse(PopTransactionStatus.BadRequest, "invalid unique-id list format"));
          }
        }
        else {
          ConcludeError(status);
        }
      }
    }

    private void RetrieveUidl()
    {
      PopProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is PopFollowingResponse) {
        PopUniqueIdList uidList;
        if (RetrieveUniqueIdList(response as PopFollowingResponse, out uidList))
          uidLists.Add(uidList);
      }
      else if (response is PopTerminationResponse) {
        Conclude(new PopUniqueIdListResponse(PopTransactionStatus.Ok,
                                             uidLists.ToArray(),
                                             "retrieved unique-id lists"));
      }
    }

    private bool RetrieveUniqueIdList(PopStatusResponse status, out PopUniqueIdList uidList)
    {
      return RetrieveUniqueIdList(status.Message, out uidList);
    }

    private bool RetrieveUniqueIdList(PopFollowingResponse following, out PopUniqueIdList uidList)
    {
      return RetrieveUniqueIdList(following.Text, out uidList);
    }

    private bool RetrieveUniqueIdList(string text, out PopUniqueIdList uidList)
    {
      uidList = null;

      var resp = text.Split(new[] {' '}, StringSplitOptions.None);

      if (resp.Length < 2)
        return false;

      int messageNumber;

      if (!int.TryParse(resp[0], out messageNumber))
        return false;

      uidList = new PopUniqueIdList(messageNumber, resp[1]);

      return true;
    }

    private bool listMultiple;
    private readonly List<PopUniqueIdList> uidLists = new List<PopUniqueIdList>();
  }
}
